@@ -13,8 +13,8 @@
 package Config::Generator::XML;
 use strict;
 use warnings;
-our $VERSION  = "0.9";
-our $REVISION = sprintf("%d.%02d", q$Revision: 1.16 $ =~ /(\d+)\.(\d+)/);
+our $VERSION  = "1.0";
+our $REVISION = sprintf("%d.%02d", q$Revision: 1.18 $ =~ /(\d+)\.(\d+)/);
 
 #
 # used modules
@@ -64,13 +64,20 @@ my @xml_element_options = (
 
 sub xml_element ($;$@) {
     my($name, $attr, @body) = validate_pos(@_, @xml_element_options,
-        ({ type => HASHREF }) x (@_ - 2),
+        @_ == 3 ? ({ type => SCALAR | HASHREF })
+                : ({ type => HASHREF }) x (@_ - 2)
     );
     my(%elt);
 
     $elt{name} = $name;
     $elt{attr} = $attr if $attr;
-    $elt{body} = \@body if @body;
+    if (@body) {
+        if (@body == 1 and not ref($body[0])) {
+            $elt{body} = $body[0];
+        } else {
+            $elt{body} = \@body;
+        }
+    }
     return(\%elt);
 }
 
@@ -103,6 +110,35 @@ sub _escape ($) {
     return($string);
 }
 
+#
+# return the list of attributes of an XML element
+#
+
+sub _attrs ($%) {
+    my($elt, %option) = @_;
+    my($eltname, @names, @attrs);
+
+    return() unless $elt->{attr};
+    $eltname = $elt->{name};
+    if (ref($elt->{attr}) eq "HASH") {
+        # given as hash for convenience
+        if ($option{sort}{$eltname}) {
+            @names = $option{sort}{$eltname}->(keys(%{ $elt->{attr} }));
+        } else {
+            @names = sort(keys(%{ $elt->{attr} }));
+        }
+        foreach my $name (@names) {
+            push(@attrs, $name . '="' . $elt->{attr}{$name} . '"');
+        }
+    } elsif (ref($elt->{attr}) eq "ARRAY") {
+        # given as name=value array for full control
+        @attrs = @{ $elt->{attr} };
+    } else {
+        dief("unexpected XML attribute: %s", $elt->{attr});
+    }
+    return(@attrs);
+}
+
 #
 # return a string representing a blank line or an XML comment
 #
@@ -133,45 +169,32 @@ sub _strspc ($$) {
 sub _strelt ($%);
 sub _strelt ($%) {
     my($elt, %option) = @_;
-    my($eltname, $indent, $xml, @names, @attrs, $line, $sep);
+    my($eltname, $indent, $xml, @attrs, $line, $sep);
 
     $eltname = $elt->{name};
     $indent = TAB() x $option{indent};
-    $option{indent}++;
     return(_strspc($elt->{body}, $indent)) unless $eltname;
     $xml = $indent . "<" . $eltname;
-    if ($elt->{attr}) {
-        if (ref($elt->{attr}) eq "HASH") {
-            # given as hash for convenience
-            if ($option{sort}{$eltname}) {
-                @names = $option{sort}{$eltname}->(keys(%{ $elt->{attr} }));
-            } else {
-                @names = sort(keys(%{ $elt->{attr} }));
-            }
-            foreach my $name (@names) {
-                push(@attrs, $name . '="' . $elt->{attr}{$name} . '"');
-            }
-        } elsif (ref($elt->{attr}) eq "ARRAY") {
-            # given as name=value array for full control
-            @attrs = @{ $elt->{attr} };
+    @attrs = _attrs($elt, %option);
+    if (@attrs) {
+        @attrs = map(_escape($_), @attrs);
+        $line = $option{line};
+        $line = 1 if $option{split}{$eltname};
+        if ($line and length("$indent<$eltname @attrs>\n") > $line) {
+            $sep = "\n$indent";
+            @attrs = (map(TAB() . $_, @attrs), "");
         } else {
-            dief("unexpected XML attribute: %s", $elt->{attr});
-        }
-        if (@attrs) {
-            @attrs = map(_escape($_), @attrs);
-            $line = $option{line};
-            $line = 1 if $option{split}{$eltname};
-            if ($line and length("$indent<$eltname @attrs>\n") > $line) {
-                $sep = "\n$indent";
-                @attrs = (map(TAB() . $_, @attrs), "");
-            } else {
-                $sep = " ";
-            }
-            $xml .= join($sep, "", @attrs);
+            $sep = " ";
         }
+        $xml .= join($sep, "", @attrs);
     }
-    if ($elt->{body} and @{ $elt->{body} }) {
+    if (defined($elt->{body}) and not ref($elt->{body})) {
+        # string body
+        $xml .= ">" . $elt->{body} . "</" . $eltname . ">\n";
+    } elsif ($elt->{body} and @{ $elt->{body} }) {
+        # structure body
         $xml .= ">\n";
+        $option{indent}++;
         foreach my $child (@{ $elt->{body} }) {
             dief("unexpected XML child: %s", $child)
                 unless ref($child) eq "HASH";
@@ -179,6 +202,7 @@ sub _strelt ($%) {
         }
         $xml .= $indent . "</" . $eltname . ">\n";
     } else {
+        # no body
         $xml .= "/>\n";
     }
     return($xml);
@@ -261,36 +285,61 @@ sub Init ($) {
 
 sub Start ($$%) {
     my($parser, $tag, %attr) = @_;
-    my($newelt);
+    my($elt);
 
-    $newelt = {
+    $elt = {
         name => $tag,
         attr => \%attr,
         body => [],
     };
     push(@{ $parser->{Stack} }, $parser->{Current});
-    push(@{ $parser->{Current} }, $newelt);
-    $parser->{Current} = $newelt->{body};
+    push(@{ $parser->{Current} }, $elt);
+    $parser->{Current} = $elt->{body};
 }
 
 sub End ($$) {
     my($parser, $tag) = @_;
+    my($elt);
 
+    $elt = $parser->{Stack}[-1][-1];
+    # FIXME: forbid multi-line body?
+    if (@{ $elt->{body} } == 0) {
+        # no body
+        delete($elt->{body});
+    } elsif (@{ $elt->{body} } == 1 and not ref($elt->{body}[0])) {
+        # string body
+        $elt->{body} = $elt->{body}[0];
+    } elsif (@{ $elt->{body} } > 1) {
+        # enforce body consistency
+        foreach my $chunk (@{ $elt->{body} }) {
+            dief("unexpected text: %s", $chunk) unless ref($chunk);
+        }
+    }
     $parser->{Current} = pop(@{ $parser->{Stack} });
 }
 
 sub Char ($$) {
     my($parser, $text) = @_;
 
-    dief("unexpected text: %s", $text) unless $text =~ /^\s*$/;
+    return if $text =~ /^\s*$/;
+    $text =~ s/^\s+//;
+    $text =~ s/\s+$//;
+    dief("unexpected multi-line text: %s", $text) if $text =~ /\n/;
+    push(@{ $parser->{Current} }, $text);
 }
 
 sub Comment ($$) {
     my($parser, $text) = @_;
+    my(@lines);
 
     $text =~ s/^\s+//;
     $text =~ s/\s+$//;
-    push(@{ $parser->{Current} }, { body => $text });
+    foreach my $line (split(/\n/, $text)) {
+        $line =~ s/^\s+//;
+        $line =~ s/\s+$//;
+        push(@lines, $line);
+    }
+    push(@{ $parser->{Current} }, { body => join("\n", @lines) });
 }
 
 sub Final ($) {
@@ -316,8 +365,8 @@ generate XML abstractions (xml_blank(), xml_comment(), xml_element(),
 xml_parse() and xml_wrap()) and one function to convert these abstractions
 into strings (xml_string()).
 
-Note: only a subset of XML (without any text) is supported. This means that
-this module is not suited for HTML generation.
+Note: only a subset of XML is supported. This means that this module is not
+suited for HTML generation.
 
 =head1 FUNCTIONS
 
@@ -342,7 +391,7 @@ return an XML abstraction representing an element:
 
 =item * ATTR: defines the attributes as a hash or array reference
 
-=item * BODY: defines the body as a list of XML abstractions
+=item * BODY: defines the body as a string or list of XML abstractions
 
 =back
 
@@ -381,4 +430,4 @@ attributes)
 
 Lionel Cons L<http://cern.ch/lionel.cons>
 
-Copyright (C) CERN 2013-2015
+Copyright (C) CERN 2013-2016
